import gzip

# 3 sets of 10 colors
COLORS = [
    "#1f77b4",  # plotly and d3
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
    "#7CB5EC",  # highcharts
    "#434348",
    "#90ED7D",
    "#F7A35C",
    "#8085E9",
    "#F15C80",
    "#E4D354",
    "#2B908F",
    "#F45B5B",
    "#91E8E1",
    "#4E79A7",  # tableau
    "#F28E2C",
    "#E15759",
    "#76B7B2",
    "#59A14F",
    "#EDC949",
    "#AF7AA1",
    "#FF9DA7",
    "#9C755F",
    "#BAB0AB",
]


def gzopen(f):
    if f.endswith(".gz"):
        return gzip.open(f, "rt")
    else:
        return open(f)


def merge_intervals(intervals):
    sorted_intervals = sorted(intervals, key=lambda i: i[0])
    merged = list()
    for higher in sorted_intervals:
        if not merged:
            merged.append(higher)
        else:
            lower = merged[-1]
            # merge bookends
            if higher[0] - lower[1] == 1:
                # update existing entry
                merged[-1] = [lower[0], higher[1], lower[2] + higher[2]]
            elif higher[0] <= lower[1]:
                upper_bound = max(lower[1], higher[1])
                # update existing entry
                merged[-1] = [lower[0], upper_bound, lower[2] + higher[2]]
            # non-overlapping
            else:
                merged.append(higher)
    return merged



import os
import re
from itertools import groupby

from .utils import gzopen

try:
    from itertools import ifilterfalse as filterfalse
except ImportError:
    from itertools import filterfalse


def parse_vcf(path, traces, exclude, regex=None, y_offset=-0.15, track_color="#444"):
    """
    parse a VCFv4.1 file, placing squares per variant
    """

    trace_name = os.path.basename(path)

    with gzopen(path) as fh:
        cleaned = filterfalse(lambda i: i[0] == "#", fh)

        info_re = None
        if regex:
            info_re = re.compile(r"%s([^;]*)" % regex)

        for chrom, entries in groupby(
            cleaned, key=lambda i: i.partition("\t")[0].lstrip("chr")
        ):
            # apply exclusions
            if exclude.findall(chrom):
                continue
            if chrom not in traces:
                continue

            trace_x = list()
            trace_y = list()
            trace_text = list()

            for line in entries:
                if line.startswith("#"):
                    continue

                toks = line.strip().split("\t")

                # not currently converting 0- and 1-based
                x = int(toks[1])

                info = toks[7].replace(";", "<br>")
                if info_re:
                    try:
                        info = info_re.findall(toks[7])[0]
                    except IndexError:
                        info = ""

                trace_x.append(x)
                trace_y.append(y_offset)
                trace_text.append(info)

            trace = dict(
                x=trace_x,
                y=trace_y,
                mode="markers",
                type="scattergl",
                name=trace_name,
                text=trace_text,
                marker=dict(
                    size=10,
                    symbol="square",
                    color=track_color,
                    line=dict(width=1, color="white"),
                ),
                tracktype="vcf",
            )
            traces[chrom].append(trace)
    return traces



"""
The aim of covviz is to highlight regions of significant (passing the
user's z-score threshold) and sustained (beyond user specified
distance) deviation from the majority of samples. Significance is
determined using z-scores for all samples at all points using median
absolute deviation, but in order to be highlighted, points must be
significant consecutively throughout a user specified distance.

If you are analyzing a low number of samples, deviation may be
irrelevant. In this case, we can set --min-samples to be greater
than our sample total to skip Z-threshold calculation and plot
coverages for all samples at all points.

Annotation tracks, --bed, --gff, and --vcf can be specified more than
once.
"""

import argparse
import csv
import json
import logging
import os
import re
from collections import defaultdict

from jinja2 import Environment, FileSystemLoader, select_autoescape
from lzstring import LZString

from .bed import parse_bed, parse_bed_track
from .gff import parse_gff
from .ped import parse_ped
from .utils import gzopen
from .vcf import parse_vcf

logging.basicConfig(level=logging.INFO, format="[%(levelname)s] %(message)s")
logger = logging.getLogger("covviz")


class Formatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
):
    pass


def parse_args():
    p = argparse.ArgumentParser(description=__doc__, formatter_class=Formatter)
    # this will later be the only requirement
    p.add_argument(
        "bed",
        help=(
            "bed3+ file format with a header defining sample columns "
            "after chrom, start, and end."
        ),
    )
    p.add_argument(
        "-e",
        "--exclude",
        default="^HLA,^hs,:,^GL,M,EBV,^NC,^phix,decoy,random$,Un,hap,_alt$",
        help="chromosome regex to exclude from analysis",
    )
    p.add_argument(
        "-x",
        "--sex-chroms",
        default="X,Y",
        help="sex chromosomes as they are defined in your bed, e.g. chrX,chrY",
    )

    p.add_argument(
        "-z",
        "--z-threshold",
        default=3.5,
        type=float,
        help=(
            "the point at which we determine a sample is an outlier "
            "from the group at any given point"
        ),
    )
    p.add_argument(
        "-d",
        "--distance-threshold",
        default=150000,
        type=int,
        help=(
            "when calculating significance, the Z-score has to be "
            "above the threshold for consecutive points up to the "
            "total distance set by distance threshold"
        ),
    )
    p.add_argument(
        "-s",
        "--slop",
        default=500000,
        type=int,
        help=(
            "slop is the distance to add to traces when plotting -- "
            "without slop, it's not always clear what happens to the "
            "points immediately flanking the area of significant "
            "deviation"
        ),
    )
    p.add_argument(
        "-o", "--output", default="covviz_report.html", help="output file path"
    )
    p.add_argument(
        "--skip-norm",
        action="store_true",
        help=(
            "skip normalization by global sample median if the depths "
            "in your .bed are already normalized"
        ),
    )
    p.add_argument(
        "--min-samples",
        default=8,
        type=int,
        help=(
            "show all traces when analyzing this few samples; ignores "
            "z-threshold, distance-threshold, and slop"
        ),
    )

    meta_group = p.add_argument_group("sample metadata")
    meta_group.add_argument(
        "-p", "--ped", help="ped file defining samples, sex, and other metadata"
    )
    meta_group.add_argument(
        "--sample-col",
        default="sample_id",
        help="when using --ped, this defines the sample ID column",
    )
    meta_group.add_argument(
        "--sex-col", default="sex", help="when using --ped, this defines the sex column"
    )
    meta_group.add_argument(
        "--sex-vals",
        default="1,2",
        help="when using --ped, this defines male,female encoding",
    )

    annotations_group = p.add_argument_group("annotations")
    annotations_group.add_argument(
        "--bed",
        dest="bed_track",
        action="append",
        help=(
            ".bed reference file to place an optional region track; "
            "the name field is displayed on hover and overlapping "
            "regions will be merged."
        ),
    )
    annotations_group.add_argument(
        "--gff",
        action="append",
        help=(
            ".gff reference file to place an optional gene track; "
            "only rows of type 'gene' are used and annotated with "
            "gene ID where the attributes include 'Name=<symbol>;' "
            "annotation"
        ),
    )
    annotations_group.add_argument(
        "--gff-feature",
        default="gene",
        help="feature type within GFF upon which to search for attribute",
    )
    annotations_group.add_argument(
        "--gff-attr",
        default="Name=",
        help="the regex search string to grab the relevant portion of the attributes",
    )
    annotations_group.add_argument(
        "--vcf",
        action="append",
        help=(
            ".vcf file to place an optional variant track; info is "
            "displayed by default, but can be broken up using "
            "--vcf-info regex"
        ),
    )
    annotations_group.add_argument(
        "--vcf-info",
        default=None,
        help=(
            "the regex search string to grab the relevant portion of "
            "the INFO field, e.g. 'CLNDN=' for the clinical diagnosis "
            "field in ClinVar"
        ),
    )

    return p.parse_args()


def compress_data(data):
    json_str = json.dumps(data).encode("utf-8", "ignore").decode("utf-8")
    json_str = json_str.replace("NaN", "null")
    return LZString().compressToBase64(json_str)


def cli():
    args = parse_args()

    env = Environment(
        loader=FileSystemLoader(os.path.join(os.path.dirname(__file__), "templates")),
        autoescape=select_autoescape(["html"]),
    )

    logger.info("parsing bed file (%s)" % args.bed)

    exclude = re.compile(args.exclude.replace("~", "").replace(",", "|"))

    traces, samples = parse_bed(
        args.bed,
        exclude,
        args.ped,
        args.sample_col,
        args.sex_col,
        args.sex_chroms,
        args.z_threshold,
        args.distance_threshold,
        args.slop,
        args.min_samples,
        args.skip_norm,
    )

    annotation_track_colors = [
        "#666666",
        "#1b9e77",
        "#d95f02",
        "#7570b3",
        "#e7298a",
        "#66a61e",
        "#e6ab02",
        "#a6761d",
    ]

    y_offset = -0.15
    track_id = 0

    if args.gff:
        for gff in args.gff:
            logger.info("parsing gff file (%s)" % gff)
            traces = parse_gff(
                gff,
                traces,
                exclude,
                ftype=args.gff_feature,
                regex=args.gff_attr,
                y_offset=y_offset,
                track_color=annotation_track_colors[
                    track_id % len(annotation_track_colors)
                ],
            )
            y_offset += -0.15
            track_id += 1

    if args.bed_track:
        for bed in args.bed_track:
            logger.info("parsing bed file (%s)" % bed)
            traces = parse_bed_track(
                bed,
                traces,
                exclude,
                y_offset=y_offset,
                track_color=annotation_track_colors[
                    track_id % len(annotation_track_colors)
                ],
            )
            y_offset += -0.15
            track_id += 1

    if args.vcf:
        for vcf in args.vcf:
            logger.info("parsing vcf file (%s)" % vcf)
            traces = parse_vcf(
                vcf,
                traces,
                exclude,
                regex=args.vcf_info,
                y_offset=y_offset,
                track_color=annotation_track_colors[
                    track_id % len(annotation_track_colors)
                ],
            )
            y_offset += -0.15
            track_id += 1

    if args.ped:
        logger.info("parsing ped file (%s)" % args.ped)
        traces = parse_ped(
            args.ped, traces, args.sample_col, args.sex_chroms, args.sex_vals
        )

    with open(args.output, "w") as fh:
        logger.info("preparing output")
        compressed_json_data = compress_data(traces)
        html_template = env.get_template("covviz.html")
        print(html_template.render(data=compressed_json_data), file=fh)

    logger.info("processing complete")



import csv
from collections import defaultdict

from .utils import gzopen


def check_ped_header(header, cols):
    """
    header (list) - current ped header columns
    cols (list) - typical columns we'll see from indexcov
    """
    header = set(header)
    cols = set(cols)
    is_indexcov = True
    if len(cols - header) > 0:
        is_indexcov = False
    return is_indexcov


def parse_ped(path, traces, sample_col, sex_chroms, sex_vals="1,2"):
    table_data = list()
    ped_data = dict(
        inferred=defaultdict(list), bins=defaultdict(list), pca=defaultdict(list)
    )
    sex_chroms = [i.strip() for i in sex_chroms.split(",")]

    male = None
    female = None
    for i, v in enumerate(sex_vals.split(",")):
        if i == 0:
            male = v
        elif i == 1:
            female = v
        else:
            break

    # indexcov .ped format
    required_vals = [
        sample_col,
        "bins.in",
        "bins.out",
        "bins.lo",
    ]  # CNX, CNY or CNchrX, CNchrY
    float_vals = ["slope", "p.out", "PC1", "PC2", "PC3", "PC4", "PC5"]
    int_vals = ["bins.out", "bins.lo", "bins.hi", "bins.in"]

    via_indexcov = False

    with gzopen(path) as fh:
        header = fh.readline().strip().split("\t")
        # datatables needs these escaped
        datatable_cols = [dict(title=i, data=i.replace(".", "\\.")) for i in header]
        table_data.append(datatable_cols)
        fh.seek(0)
        reader = csv.DictReader(fh, delimiter="\t")

        via_indexcov = check_ped_header(header, required_vals)

        for row in reader:

            # coerce vals
            if via_indexcov:
                for k, v in row.items():
                    if k in int_vals:
                        row[k] = int(v)
                    elif k in float_vals:
                        row[k] = float(v)
                    elif k.startswith("CN"):
                        row[k] = float(v)

            # coerced vals saved in data.ped
            table_data.append(row)

            if via_indexcov:
                # inferred sex
                try:
                    ped_data["inferred"]["x"].append(row["CN%s" % sex_chroms[0]])
                except KeyError:
                    ped_data["inferred"]["x"].append(row["CNchr%s" % sex_chroms[0]])
                try:
                    ped_data["inferred"]["y"].append(row["CN%s" % sex_chroms[1]])
                except KeyError:
                    try:
                        ped_data["inferred"]["y"].append(row["CNchr%s" % sex_chroms[1]])
                    except IndexError:
                        ped_data["inferred"]["y"].append(0)
                except IndexError:
                    ped_data["inferred"]["y"].append(0)

                # male
                if row["sex"] == male:
                    ped_data["inferred"]["color"].append("rgba(144,237,125,0.5)")
                    ped_data["inferred"]["hover"].append(
                        "Sample: %s<br>Inferred X CN: 1" % (row[sample_col],)
                    )
                # female
                elif row["sex"] == female:
                    ped_data["inferred"]["color"].append("rgba(247,163,92,0.5)")
                    ped_data["inferred"]["hover"].append(
                        "Sample: %s<br>Inferred X CN: 2" % (row[sample_col],)
                    )
                else:
                    ped_data["inferred"]["color"].append("rgba(105,105,105,0.5)")
                    ped_data["inferred"]["hover"].append(
                        "Sample: %s<br>Unknown" % (row[sample_col],)
                    )

                # bin plot
                total = row["bins.in"] + row["bins.out"]
                ped_data["bins"]["samples"].append(row[sample_col])
                ped_data["bins"]["x"].append(row["bins.lo"] / total)
                ped_data["bins"]["y"].append(row["bins.out"] / total)

                # PCAs
                try:
                    ped_data["pca"]["pca_1"].append(row["PC1"])
                    ped_data["pca"]["pca_2"].append(row["PC2"])
                    ped_data["pca"]["pca_3"].append(row["PC3"])
                except KeyError:
                    pass

    traces["ped"] = table_data
    traces["sample_column"] = sample_col
    if via_indexcov:
        traces["depth"] = ped_data
    return traces



import os
import re
from itertools import groupby

from .utils import gzopen, merge_intervals

try:
    from itertools import ifilterfalse as filterfalse
except ImportError:
    from itertools import filterfalse


def parse_gff(
    path,
    traces,
    exclude,
    ftype="gene",
    regex="Name=",
    y_offset=-0.15,
    track_color="#444",
):
    """
    Grabs the gene name from the attrs field where 'Name=<symbol>;' is present.

    returns:
        dict of lists
    """
    trace_name = os.path.basename(path)
    gene_search = list()
    with gzopen(path) as fh:
        cleaned = filterfalse(lambda i: i[0] == "#", fh)
        name_re = re.compile(r"%s([^;]*)" % regex)
        for chr, entries in groupby(
            cleaned, key=lambda i: i.partition("\t")[0].lstrip("chr")
        ):
            # apply exclusions
            if exclude.findall(chr):
                continue
            if chr not in traces:
                continue
            genes = list()
            for line in entries:
                if line.startswith("#"):
                    continue
                toks = line.strip().split("\t")
                if toks[2] != ftype:
                    continue
                # not currently converting 0- and 1-based
                start = int(toks[3])
                end = int(toks[4])
                try:
                    name = name_re.findall(toks[8])[0]
                    name = name.strip('"').strip("'")
                except IndexError:
                    name = ""
                genes.append([start, end, [name]])
                gene_search.append(dict(n=name, v=[chr, start, end]))
            if genes:
                # merge overlapping genes
                merged_genes = merge_intervals(genes)
                # update gene lists to semi-colon delimited string
                for interval in merged_genes:
                    interval[2] = ";".join(set(interval[2]))
                # make the trace structure for the gene annotation track
                x_values = list()
                y_values = list()
                text_values = list()
                for interval in merged_genes:
                    # start
                    x_values.append(interval[0])
                    # end
                    x_values.append(interval[1])
                    # gap
                    x_values.append("")
                    y_values.append(y_offset)
                    y_values.append(y_offset)
                    y_values.append("")
                    text_values.append(interval[2])
                    text_values.append(interval[2])
                    text_values.append("")
                gene_trace = dict(
                    x=x_values,
                    y=y_values,
                    text=text_values,
                    type="scattergl",
                    name=trace_name,
                    tracktype="gff",
                    connectgaps=False,
                    # hoverinfo="text",
                    showlegend=False,
                    line={"width": 10, "color": track_color},
                )
                traces[chr].append(gene_trace)
    traces["genes"] = gene_search
    return traces
